/* eslint-disable */
/**
 * Original sources:
 *  - https://github.com/kmalakoff/response-iterator/blob/master/src/index.ts
 *  - https://github.com/apollographql/apollo-client/blob/main/src/link/http/responseIterator.ts
 */

import { Buffer } from "node:buffer";
import { Readable } from "node:stream";

const canUseSymbol =
  typeof Symbol === "function" && typeof Symbol.for === "function";

const canUseAsyncIteratorSymbol = canUseSymbol && Symbol.asyncIterator;

function isNodeResponse(value) {
  return !!value.body;
}

function isReadableStream(value) {
  return !!value.getReader;
}

function isAsyncIterableIterator(value) {
  return !!(canUseAsyncIteratorSymbol && value[Symbol.asyncIterator]);
}

function isStreamableBlob(value) {
  return !!value.stream;
}

function isBlob(value) {
  return !!value.arrayBuffer;
}

function isNodeReadableStream(value) {
  return !!value.pipe;
}

function asyncIterator(source) {
  let _a;
  const iterator = source[Symbol.asyncIterator]();
  return (
    (_a = {
      next() {
        return iterator.next();
      },
    }),
    (_a[Symbol.asyncIterator] = function () {
      return this;
    }),
    _a
  );
}

function readerIterator(reader) {
  const iterator = {
    next() {
      return reader.read();
    },
  };
  if (canUseAsyncIteratorSymbol) {
    iterator[Symbol.asyncIterator] = function () {
      return this;
    };
  }
  return iterator;
}

function promiseIterator(promise) {
  let resolved = false;
  const iterator = {
    next() {
      if (resolved)
        return Promise.resolve({
          value: undefined,
          done: true,
        });
      resolved = true;
      return new Promise(function (resolve, reject) {
        promise
          .then(function (value) {
            resolve({ value, done: false });
          })
          .catch(reject);
      });
    },
  };
  if (canUseAsyncIteratorSymbol) {
    iterator[Symbol.asyncIterator] = function () {
      return this;
    };
  }
  return iterator;
}

function nodeStreamIterator(stream) {
  let cleanup = null;
  let error = null;
  let done = false;
  const data = [];
  const waiting = [];
  function onData(chunk) {
    if (error) return;
    if (waiting.length) {
      const shiftedArr = waiting.shift();
      if (Array.isArray(shiftedArr) && shiftedArr[0]) {
        return shiftedArr[0]({ value: chunk, done: false });
      }
    }
    data.push(chunk);
  }
  function onError(err) {
    error = err;
    const all = waiting.slice();
    all.forEach(function (pair) {
      pair[1](err);
    });
    !cleanup || cleanup();
  }
  function onEnd() {
    done = true;
    const all = waiting.slice();
    all.forEach(function (pair) {
      pair[0]({ value: undefined, done: true });
    });
    !cleanup || cleanup();
  }
  cleanup = function () {
    cleanup = null;
    stream.removeListener("data", onData);
    stream.removeListener("error", onError);
    stream.removeListener("end", onEnd);
    stream.removeListener("finish", onEnd);
    stream.removeListener("close", onEnd);
  };
  stream.on("data", onData);
  stream.on("error", onError);
  stream.on("end", onEnd);
  stream.on("finish", onEnd);
  stream.on("close", onEnd);
  function getNext() {
    return new Promise(function (resolve, reject) {
      if (error) return reject(error);
      if (data.length) return resolve({ value: data.shift(), done: false });
      if (done) return resolve({ value: undefined, done: true });
      waiting.push([resolve, reject]);
    });
  }
  const iterator = {
    next() {
      return getNext();
    },
  };
  if (canUseAsyncIteratorSymbol) {
    iterator[Symbol.asyncIterator] = function () {
      return this;
    };
  }
  return iterator;
}

export function responseIterator(response) {
  let body = response;
  if (isNodeResponse(response)) body = response.body;
  if (Buffer.isBuffer(response)) body = Readable.from(response);
  if (isAsyncIterableIterator(body)) return asyncIterator(body);
  if (isReadableStream(body)) return readerIterator(body.getReader());
  if (isStreamableBlob(body)) return readerIterator(body.stream().getReader());
  if (isBlob(body)) return promiseIterator(body.arrayBuffer());
  if (isNodeReadableStream(body)) return nodeStreamIterator(body);
  throw new Error(
    "Unknown body type for responseIterator. Please pass a streamable response."
  );
}
